indexing

	description: "Class that covers the Standard C stream%
	%code. Filenames are subject to expansion."

	idea: "Mimicks an existing file. Tries to maintain that file exists%
	%during life of object by keeping it open from creation to%
	%destruction of the object. But you may close it earlier."

	author: "Berend de Boer"
	date: "$Date: 2003/01/09 $"
	revision: "$Revision: #3 $"


deferred class

	STDC_FILE

inherit

	STDC_HANDLE [POINTER]
		rename
			handle as stream
		redefine
			do_close
		end

	EPX_POINTER_ADD

	PORTABLE_PATH

	CAPI_STDIO


feature -- Initialization

	create_read_write (path: STRING) is
			-- Open file for update (reading and writing). If the file
			-- already exists, it is truncated to zero length.
			-- So permissions seem to remain.
		require
			closed: not is_open
		do
			do_make
			do_open (path, "w+", True)
		ensure
			file_is_open: security.error_handling.exceptions_enabled implies is_open
			owner_set: is_open implies is_owner
			open_files_increased_by_one:
				is_owner implies
					security.files.open_files = old security.files.open_files + 1
		end

	create_write (path: STRING) is
			-- create new file for writing. If the file already exists,
			-- it is truncated to zero length.
			-- So permissions seem to remain.
		require
			closed: not is_open
		do
			do_make
			do_open (path, "w", True)
		ensure
			file_is_open: security.error_handling.exceptions_enabled implies is_open
			owner_set: is_open implies is_owner
			open_files_increased_by_one:
				is_owner implies
					security.files.open_files = old security.files.open_files + 1
		end

	open (path, a_mode: STRING) is
			-- open file in given mode
		require
			closed: not is_open
		do
			do_make
			do_open (path, a_mode, True)
		ensure
			file_is_open: security.error_handling.exceptions_enabled implies is_open
			owner_set: is_open implies is_owner
			open_files_increased_by_one:
				is_owner implies
					security.files.open_files = old security.files.open_files + 1
		end

	open_append (path: STRING) is
			-- append to exiting file or create file if it does not exist
		require
			closed: not is_open
		do
			do_make
			open (path, "a")
		ensure
			file_is_open: security.error_handling.exceptions_enabled implies is_open
			owner_set: is_open implies is_owner
			open_files_increased_by_one:
				is_owner implies
					security.files.open_files = old security.files.open_files + 1
		end

	open_read (path: STRING) is
			-- open file for reading
		require
			closed: not is_open
		do
			do_make
			open (path, "r")
		ensure
			file_is_open: security.error_handling.exceptions_enabled implies is_open
			owner_set: is_open implies is_owner
			open_files_increased_by_one:
				is_owner implies
					security.files.open_files = old security.files.open_files + 1
		end

	open_read_write (path: STRING) is
			-- Open file for reading and writing.
		require
			closed: not is_open
		do
			do_make
			open (path, "r+")
		ensure
			file_is_open: security.error_handling.exceptions_enabled implies is_open
			owner_set: is_open implies is_owner
			open_files_increased_by_one:
				is_owner implies
					security.files.open_files = old security.files.open_files + 1
		end


feature -- Work with existing streams

	attach_to_stream (a_stream: POINTER; a_mode: STRING) is
			-- Attach to `a_stream'. Does not become owner of stream so
			-- it will not close on `close' or when garbage collected.
		require
			closed: not is_open
			valid_stream: a_stream /= Void
			valid_mode: a_mode /= Void and then a_mode.count > 0
			-- `a_stream' is open
			-- `a_mode' is compatible with `a_stream'
		do
			do_make
			set_mode (a_mode)
			capacity := 1
			set_handle (a_stream, False)
		ensure
			opened: is_open
			not_owner: not is_owner
			open_files_unchanged:
				security.files.open_files = old security.files.open_files
		end


feature -- Close

	unattach is
		obsolete "Use detach instead"
		do
			detach
		end


feature -- Ownership of the stream. Can help to influence subtile garbage collector problems

	make_owner is
		obsolete "Use become_owner instead."
		do
			become_owner
		end


feature -- Reopen

	reopen (path, a_mode: STRING) is
			-- Closes and then opens a stream.
		require
			open: is_open
			-- valid_mode: mode is a valid posix mode
		local
			cpath: POINTER
			cmode: POINTER
		do
			set_portable_path (path)
			cpath := sh.string_to_pointer (portable_path)
			set_mode (a_mode)
			cmode := sh.string_to_pointer (mode)
			stream := posix_freopen (cpath, cmode, stream)
			if not is_open then
				raise_posix_error
			end
		ensure
			file_stays_open: is_open
			open_files_unchanged:
				security.files.open_files = old security.files.open_files
		end


feature -- Control over buffering

	flush is
			-- Updates this stream
		require
			open: is_open
		local
			i: INTEGER
		do
			i := posix_fflush (stream)
			if i /= 0 then
				raise_posix_error
			end
		end

	setbuf, set_buffer (buffer: POINTER) is
			-- Determines how the stream will be buffered
			-- gives you a fully buffered input and output.
			-- Not sure: buffer should have at least BUFSIZ bytes?
			-- No operation should yet been performed on this file
			-- `buffer' = default_pointer: default buffer will be allocated
			-- `buffer' /= default_pointer implies buffer size = BUFSIZ
		require
			open: is_open
		do
			posix_setbuf (stream, buffer)
		end

	set_full_buffering (buffer: POINTER; size: INTEGER) is
			-- Determines buffering for a stream.
			-- If `buffer' is default_pointer, a buffer of `size' bytes
			-- will be allocated by this routine.
		require
			open: is_open
		local
			r: INTEGER
		do
			r := posix_setvbuf (stream, buffer, IOFBF, size)
			if r /= 0 then
				raise_posix_error
			end
		end

	set_line_buffering (buffer: POINTER; size: INTEGER) is
			-- Determines buffering for a stream.
			-- Give NULL `buffer' so `setvbuf' will allocate a buffer.
		require
			open: is_open
		local
			r: INTEGER
		do
			r := posix_setvbuf (stream, buffer, IOLBF, size)
			if r /= 0 then
				raise_posix_error
			end
		end

	set_no_buffering is
			-- Turn buffering off.
		require
			open: is_open
		local
			r: INTEGER
		do
			r := posix_setvbuf (stream, default_pointer, IONBF, 0)
			if r /= 0 then
				raise_posix_error
			end
		end


feature -- read, C like

	last_byte: INTEGER
			-- last read character of `get_character'
			-- can be negative, so is more a last_shortint or so!

	getc, get_character is
			-- Reads a C unsigned char and converts it to an integer,
			-- the result is left in `last_byte'
			-- This function probably can be used to read a single
			-- byte
		require
			open: is_open
		do
			assert_input_ok
			last_byte := posix_fgetc (stream)
			eof_read := last_byte = const_EOF
		ensure
			eof_set: last_byte = const_EOF implies eof
		end

	gets, get_string (bytes: INTEGER) is
			-- Reads at most one less than `bytes' characters.
			-- No additional characters are read after a newline character
			-- or after end-of-file. If a newline character is read, it
			-- is returned too.
			-- Result is placed in `last_string'
		require
			open: is_open
		local
			p: POINTER
		do
			assert_input_ok
			assert_buf_has_room (bytes)
			p := posix_fgets (gets_buf.ptr, bytes, stream)
			if p = default_pointer then
				eof_read := posix_feof (stream)
				if eof_read or else bytes <= 1 then
					last_string.wipe_out
				else
					raise_posix_error
				end
			else
				sh.set_string_from_pointer (last_string, gets_buf.ptr)
				if eof then
					-- we don't want the eof to be True here, user has to call
					-- gets again to make Eof True, fits better with Eiffel loops
					clear_error
				end
			end
			last_read := last_string.count
		ensure
			valid_string: last_string.count <= bytes
			read_consistent: last_read <= bytes
		end

	read (buf: POINTER; offset, bytes: INTEGER) is
			-- Read chunk, set `last_read'. `offset' determines how far
			-- in `buf' you want to start writing.
		require
			open: is_open
			valid_buf: buf /= default_pointer
			valid_bytes: bytes >= 0
		local
			ptr: POINTER
		do
			assert_input_ok
			ptr := posix_pointer_add (buf, offset)
			last_read := posix_fread (ptr, 1, bytes, stream)
			if eof then
				if last_read > 0 then
					-- we don't want the eof to be True when bytes are read.
					-- user has to call read again to make Eof True, fits
					-- better with Eiffel loops
					clear_error
				end
			else
				-- not eof, but not all requested bytes read?? Stop!
				if  last_read /= bytes then
					raise_posix_error
				end
			end
		ensure
			successfull_read: not eof implies last_read > 0
			read_consistent: last_read <= bytes
		end


feature {NONE} -- gets private state

	eof_read: BOOLEAN
			-- set by `getc' and `gets', signals EOF read

	gets_buf: STDC_BUFFER
			-- buffer used by `gets'

	assert_buf_has_room (needed_size: INTEGER) is
			-- make sure `gets_buf' has enough room
		do
			if gets_buf.capacity < needed_size then
				gets_buf.resize (needed_size)
			end
		end


feature -- write, C like

	putc, put_character (c: INTEGER) is
			-- write a single character
		require
			open: is_open
		local
			r: INTEGER
		do
			r := posix_fputc (c, stream)
			if r = const_EOF then
				raise_posix_error
			end
			need_flush := True
		ensure
			need_flush_set: need_flush
		end

	ungetc (c: INTEGER) is
			-- pushes `c' back to the stream
			-- note that file positioning functions discard any
			-- pushed-back characters
		require
			open: is_open
		local
			r: INTEGER
		do
			-- and that's why we make sure position is not needed anymore
			assert_input_ok
			r := posix_ungetc (c, stream)
			if r = const_EOF then
				raise_posix_error
			end
		end

	write (buf: POINTER; offset, bytes: INTEGER) is
			-- write `bytes' bytes from `buf' at offset `offset'
			-- we do not really care if offset is positive or negative...
		require
			open: is_open
			valid_buf: buf /= default_pointer
			valid_bytes: bytes >= 0
		local
			ptr: POINTER
		do
			ptr := posix_pointer_add (buf, offset)
			last_written := posix_fwrite (ptr, 1, bytes, stream)
			if last_written /= bytes then
				raise_posix_error
			end
			need_flush := True
		ensure
			successfull_write: last_written = bytes
			need_flush_set: need_flush
		end


feature -- read, Eiffel like

	last_read: INTEGER
			-- last read bytes by some read_XXXX or get_string call

	last_boolean: BOOLEAN
			-- last boolean read by `read_boolean'

	last_character: CHARACTER
			-- last character read by `read_character'

	last_double: DOUBLE
			-- last double lread by `read_double'

	last_integer: INTEGER

	last_real: REAL
			-- last real read by `read_real'

	last_string: STRING
			-- Last string read by `read_string' or
			-- `get_string'. Includes the end-of-line character, if any.

	read_boolean is
			-- attempt to read back a boolean written by write_boolean
		require
			open: is_open
		do
			read_character
			inspect last_character
			when 't', 'T' then
				read_string (3)
				if last_string.is_equal ("rue") then
					last_boolean := true
				else
					-- other ideas?
					raise_posix_error
				end
			when 'f', 'F' then
				read_string (4)
				if last_string.is_equal ("alse") then
					last_boolean := false
				else
					-- set errno here??
					raise_posix_error
				end
			else
				-- set errno here??
				raise_posix_error
			end
		end

	read_buffer (buf: STDC_BUFFER; offset, bytes: INTEGER) is
			-- More safe version of `read' in case you have a
			-- STDC_BUFFER object. Read starts at `offset' bytes in `buf'.
			-- Check `last_read' for number of bytes actually read.
		require
			open: is_open
			offset_not_negative: offset >= 0
			dont_overflow: buf.is_valid_index (offset + bytes - 1)
			valid_bytes: bytes >= 0
		do
			read (buf.ptr, offset, bytes)
		end

	read_double is
		require
			open: is_open
		do
			assert_input_ok
			check_fscanf (posix_fscanf_double (stream, $last_double))
		end

	read_character is
			-- read a single character and set `last_character'
			-- if end-of-file encountered, `eof' is True
		require
			open: is_open
		do
			get_character
			last_character := ch.from_integer (last_byte)
			last_read := 1
		end

	read_integer is
		require
			open: is_open
		do
			assert_input_ok
			check_fscanf (posix_fscanf_integer (stream, $last_integer))
		end

	read_real is
		require
			open: is_open
		do
			assert_input_ok
			check_fscanf (posix_fscanf_real (stream, $last_real))
		end

	read_string (bytes: INTEGER) is
			-- Read at most `n' characters, a value more expected by
			-- programmers not used to strings with a trailing byte.
			-- result is placed in `last_string'
			-- `last_string' includes the newline character if `bytes'
			-- are longer then the length of the actual line!
		require
			open: is_open
		do
			get_string (bytes+1)
		end


feature -- write, Eiffel like

	last_written: INTEGER
			-- last written bytes by some write_XXXX call

	put (any: ANY) is
			-- write class as string
		require
			open: is_open
		do
			if any /= Void then
				write_string (any.out)
			end
		ensure
			successfull_write: last_written = any.out.count
			need_flush_set: need_flush
		end

	write_buffer (buf: STDC_BUFFER; offset, bytes: INTEGER) is
			-- more safe version of `write' in case you have a
			-- STDC_BUFFER object
			-- Check `last_written' for number of bytes actually written,
			-- if you use asynchronous writing.
		require
			open: is_open
			dont_write_garbage: buf.is_valid_index (offset + bytes - 1)
		do
			write (buf.ptr, offset, bytes)
		ensure
			successfull_write: last_written = bytes
			need_flush_set: need_flush
		end

	write_boolean (b: BOOLEAN) is
			-- write a boolean in Standard C %f format
		require
			open: is_open
		do
			write_string (b.out)
		ensure
			successfull_write: last_written = b.out.count
			need_flush_set: need_flush
		end

	write_character (c: CHARACTER) is
			-- write a single character
		require
			open: is_open
		do
			putc (c.code)
			last_written := 1
		ensure
			successfull_write: last_written = 1
			need_flush_set: need_flush
		end

	write_double (d: DOUBLE) is
			-- write a double in Standard C %f format
		require
			open: is_open
		do
			check_fprintf (posix_fprintf_double (stream, d))
		ensure
			need_flush_set: need_flush
		end

	write_integer (i: INTEGER) is
			-- write an integer in Standard C %d format
		require
			open: is_open
		do
			check_fprintf (posix_fprintf_int (stream, i))
		ensure
			need_flush_set: need_flush
		end

	write_real (r: REAL) is
			-- write a real in Standard C %f format
		require
			open: is_open
		do
			check_fprintf (posix_fprintf_real (stream, r))
		ensure
			need_flush_set: need_flush
		end

	write_string, puts, put_string (s: STRING) is
			-- write a string
		require
			open: is_open
		local
			cs: POINTER
			r: INTEGER
		do
			if s /= Void then
				cs := sh.string_to_pointer (s)
				r := posix_fputs (cs, stream)
				if r = const_EOF then
					raise_posix_error
				end
				last_written := s.count
				need_flush := True
			else
				last_written := 0
			end
		ensure
			successfull_write: (s = Void implies last_written = 0) or else
				(s /= Void implies last_written = s.count)
				need_flush_set: s /= Void implies need_flush
		end


feature -- file position

	getpos, get_position: STDC_FILE_POSITION is
			-- get the current position, use `set_position' to return to
			-- this saved position
		require
			open: is_open
		do
			create Result.make
			if posix_fgetpos (stream, Result.buf.ptr) /= 0 then
				raise_posix_error
			end
		end

	rewind is
			-- Sets the file position to the beginning of the file
		require
			open: is_open
		do
			posix_rewind (stream)
			eof_read := False
			need_flush := False
		ensure
			not_eof: not eof
			no_need_to_flush: not need_flush
		end

	seek (offset: INTEGER) is
			-- set file position to given absolute `offset'
		require
			open: is_open
			valid_offset: offset >= 0
		do
			if posix_fseek (stream, offset, SEEK_SET) /= 0 then
				raise_posix_error
			end
			eof_read := False
			need_flush := False
		ensure
			not_eof: not eof
			no_need_to_flush: not need_flush
		end

	seek_from_current (offset: INTEGER) is
			-- set file position relative to current position
		require
			open: is_open
		do
			if posix_fseek (stream, offset, SEEK_CUR) /= 0 then
				raise_posix_error
			end
			eof_read := False
			need_flush := False
		ensure
			not_eof: not eof
			no_need_to_flush: not need_flush
		end

	seek_from_end (offset: INTEGER) is
			-- set file position relative to end of file
		require
			open: is_open
			valid_offset: offset <= 0
		do
			if posix_fseek (stream, offset, SEEK_END) /= 0 then
				raise_posix_error
			end
			eof_read := False
			need_flush := False
		ensure
			not_eof: not eof
			no_need_to_flush: not need_flush
		end

	setpos, set_position (a_position: STDC_FILE_POSITION) is
			-- set the current position
		require
			open: is_open
			valid_position: a_position /= Void
		do
			if posix_fsetpos (stream, a_position.buf.ptr) /= 0 then
				raise_posix_error
			end
			eof_read := False
			need_flush := False
		ensure
			not_eof: not eof
			no_need_to_flush: not need_flush
		end

	tell: INTEGER is
			-- The current position
		require
			open: is_open
		do
			Result := posix_ftell (stream)
		end


feature -- other

	clearerr, clear_error is
			-- Clears end-of-file and error indicators for a stream.
		require
			open: is_open
		do
			posix_clearerr (stream)
			eof_read := False
		ensure
			eof_not_set: not eof
		end


feature -- Access

	eof: BOOLEAN is
			-- True if eof encountered by getc or,
			-- if the end-of-file indicator is set.
		do
			Result := eof_read or else posix_feof (stream)
		end

	error: BOOLEAN is
			-- True if and only if the error indicator is set
		require
			open: is_open
		do
			Result := posix_ferror (stream)
		end

	filename: STRING is
			--The filename of this file.
		do
			Result := portable_path
		end

	mode: STRING
			-- mode in which the file is opened/created

	resource_usage_can_be_increased: BOOLEAN is
			-- Is it allowed to open another file?
		do
			Result := security.files.open_files <= security.files.max_open_files
		end


feature -- is mode binary or text

	is_binary_mode_specification (a_mode: STRING): BOOLEAN is
			-- True if last character of a_mode = 'b'
		local
			last: CHARACTER
		do
			if a_mode.count > 0 then
				last := a_mode.item (a_mode.count)
				Result :=
					last_character = 'b' or else last_character = 'B'
			end
		end

	is_text_mode_specification (a_mode: STRING): BOOLEAN is
			-- True if last character of a_mode = 't'
		local
			last: CHARACTER
		do
			if a_mode.count > 0 then
				last := a_mode.item (a_mode.count)
				Result :=
					last_character = 't' or else last_character = 'T'
			end
		end


feature {NONE} -- internal

	check_fprintf (res: INTEGER) is
			-- Make sure the result of a fprintf is correctly handled.
		do
			if res < 0 then
				raise_posix_error
			end
			need_flush := True
			last_written := res
		end

	check_fscanf (res: INTEGER) is
			-- Make sure the result of a fscanf is correctly handled.
		do
			if res < 0 then
				raise_posix_error
			end
		end

	do_make is
			-- Make sure invariants are satisfied.
		do
			make_path
			if last_string = Void then
				last_string := sh.make_with_capacity (256)
			end
			if gets_buf = Void then
				create gets_buf.allocate_and_clear (256)
			end
		end

	do_open (a_path, a_mode: STRING; a_owner: BOOLEAN) is
			-- Internal open of stream, main entry point for all routines
			-- that actually open the stream.
		require
			valid_mode: a_mode /= Void and then not a_mode.is_empty
		local
			cpath: POINTER
			cmode: POINTER
			a_stream: like stream
		do
			set_portable_path (a_path)
			cpath := sh.string_to_pointer (portable_path)
			set_mode (a_mode)
			cmode := sh.string_to_pointer (mode)
			a_stream := posix_fopen (cpath, cmode)
			if a_stream = default_pointer then
				raise_posix_error
			else
				capacity := 1
				set_handle (a_stream, a_owner)
			end
		ensure
			file_is_open: security.error_handling.exceptions_enabled implies is_open
			owner_set: is_open implies is_owner = a_owner
			open_files_increased_by_one:
				is_owner implies
					security.files.open_files = old security.files.open_files + 1
		end

	need_flush: BOOLEAN
			-- need to be set after any write operation
			-- reset after any flush operation

	assert_input_ok is
			-- ANSI C requires that after a write, a file positioning function
			-- or flush is performed.
		require
			open: need_flush implies is_open
		do
			if need_flush then
				seek_from_current (0)
			end
		ensure
			no_need_to_flush: not need_flush
		end

	set_mode (a_mode: STRING) is
			-- use this routine to set the mode variable. It takes care
			-- of adding the binary or text flag in its descendents
		require
			valid_mode: a_mode /= Void and then not a_mode.is_empty
		deferred
		end


feature -- Low level handle functions

	do_close: BOOLEAN is
			-- Close resource. Return False if an error occurred. Error
			-- value should be in `errno'. This routine may never call
			-- another object, else it cannot be used safely in
			-- `dispose'.
			-- This routine is usely redefined to actually close or
			-- deallocate the resource in addition of resetting `handle'.
		local
			r: INTEGER
		do
			r := posix_fclose (stream)
			Result := r = 0
			clear_handle
		end

	frozen unassigned_value: POINTER is
			-- The value that indicates that `handle' is unassigned.
		do
			Result := default_pointer
		end


feature {NONE} -- Counting of allocated resource

	decrease_open_files is
			-- Decreate the number of open files, does not access any
			-- other object, so can be called from `dispose' if one
			-- inherits from this class.
		external "C"
		alias "posix_decrease_open_files"
		end

	decrease_resource_count is
			-- Decrease number of allocated resource by `capacity'.
			-- Due to limitations of certain Eiffel implementations, this
			-- routine may not call any other object. Calling C code is safe.
		do
			decrease_open_files
		end

	increase_resource_count is
			-- Increase number of allocated resources by `capacity'.
		do
			security.files.increase_open_files
		end

	resource_count: INTEGER is
			-- Currently allocated number of resources. It's a global
			-- number, counting the `capacity' of all owned resources of
			-- this type.
		do
			Result := security.files.open_files
		end


invariant

	last_string_valid: last_string /= Void
	gets_buf_valid: gets_buf /= Void

end
